--- Guilder: Monitor Module
-- This module allows the user to check on the status of their consumables. If the user is an officer or raid leader/assist, they can also query other users of the addon in the group for their status.
-- @class file
-- @name Modules/Monitor.lua
-- @release 1.1.0

if select(6, GetAddOnInfo("Guilder_" .. (debugstack():match("[o%.][d%.][u%.]les\\(.-)\\") or ""))) ~= "MISSING" then return end

local Guilder = _G.Guilder
if not Guilder then
  error("Guilder_Monitor requires Guilder")
end
local L = Guilder.L

local DEBUG = Guilder.DEBUG
local expect = Guilder.expect

-- default settings for the addon -------------------------------------------

local DATABASE_DEFAULTS = {
	["checkType"] = "Trash",
	["checkFlasks"] = true,
	["checkFood"] = true,
	["checkElixirs"] = true,
	["checkExpiring"] = true,
	["checkDead"] = false,
	["checkOffline"] = false,
	["minFlask"] = 0,
	["minElixir"] = 0,
	["minFood"] = 0,
	["reportOnReadyCheck"] = true,
	["reportOnCombatEnter"] = false,
	["reportOnCombatLeave"] = false,
	["expiryLimit"] = 10,
}

-- declare the module -------------------------------------------------------

local Guilder_Monitor = Guilder:NewModule("Monitor", "AceEvent-3.0", "AceHook-3.0")
Guilder_Monitor:SetName(L["Monitor"])
Guilder_Monitor:SetDescription(L["Check for consumables and report dead or offline people while in a group"])
Guilder_Monitor:SetDefaults(DATABASE_DEFAULTS)
if DEBUG then
	Guilder_Monitor:SetIcon([[Interface\AddOns\Guilder\Modules\Monitor\Monitor]])
else
	Guilder_Monitor:SetIcon([[Interface\AddOns\Guilder_Monitor\Monitor]])
end

-- variables ----------------------------------------------------------------

local function SpellName(spellID)
	local name = GetSpellInfo(spellID)
	return name
end

local TBC_Flasks = {
	SpellName(17626), -- Flask of the Titans
	SpellName(17627), -- [Flask of] Distilled Wisdom
	SpellName(17628), -- [Flask of] Supreme Power
	SpellName(17629), -- [Flask of] Chromatic Resistance
	SpellName(28518), -- Flask of Fortification
	SpellName(28519), -- Flask of Mighty Restoration
	SpellName(28520), -- Flask of Relentless Assault
	SpellName(28521), -- Flask of Blinding Light
	SpellName(28540), -- Flask of Pure Death
	SpellName(33053), -- Mr. Pinchy's Blessing
	SpellName(42735), -- [Flask of] Chromatic Wonder
	SpellName(40567), -- Unstable Flask of the Bandit
	SpellName(40568), -- Unstable Flask of the Elder
	SpellName(40572), -- Unstable Flask of the Beast
	SpellName(40573), -- Unstable Flask of the Physician
	SpellName(40575), -- Unstable Flask of the Soldier
	SpellName(40576), -- Unstable Flask of the Sorcerer
	SpellName(41608), -- Relentless Assault of Shattrath
	SpellName(41609), -- Fortification of Shattrath
	SpellName(41610), -- Mighty Restoration of Shattrath
	SpellName(41611), -- Supreme Power of Shattrath
	SpellName(46837), -- Pure Death of Shattrath
	SpellName(46839), -- Blinding Light of Shattrath
	SpellName(67019), -- Flask of the North (WotLK 3.2)
}

local WLK_Flasks = {
	SpellName(53755), -- Flask of the Frost Wyrm
	SpellName(53758), -- Flask of Stoneblood
	SpellName(54212), -- Flask of Pure Mojo
	SpellName(53760), -- Flask of Endless Rage
	SpellName(62380), -- Lesser Flask of Resistance  -- pathetic flask
}

local TBC_BElixirs = {
	SpellName(11390),-- Arcane Elixir
	SpellName(17538),-- Elixir of the Mongoose
	SpellName(17539),-- Greater Arcane Elixir
	SpellName(28490),-- Major Strength
	SpellName(28491),-- Healing Power
	SpellName(28493),-- Major Frost Power
	SpellName(54494),-- Major Agility
	SpellName(28501),-- Major Firepower
	SpellName(28503),-- Major Shadow Power
	SpellName(38954),-- Fel Strength Elixir
	SpellName(33720),-- Onslaught Elixir
	SpellName(54452),-- Adept's Elixir
	SpellName(33726),-- Elixir of Mastery
	SpellName(26276),-- Elixir of Greater Firepower
	SpellName(45373),-- Bloodberry - only works on Sunwell Plateau
}
local TBC_GElixirs = {
	SpellName(11348),-- Greater Armor/Elixir of Superior Defense
	SpellName(11396),-- Greater Intellect
	SpellName(24363),-- Mana Regeneration/Mageblood Potion
	SpellName(28502),-- Major Armor/Elixir of Major Defense
	SpellName(28509),-- Greater Mana Regeneration/Elixir of Major Mageblood
	SpellName(28514),-- Empowerment
	SpellName(29626),-- Earthen Elixir
	SpellName(39625),-- Elixir of Major Fortitude
	SpellName(39627),-- Elixir of Draenic Wisdom
	SpellName(39628),-- Elixir of Ironskin
}

local WLK_BElixirs = {
	SpellName(28497), -- Mighty Agility
	SpellName(53748), -- Mighty Strength
	SpellName(53749), -- Guru's Elixir
	SpellName(33721), -- Spellpower Elixir
	SpellName(53746), -- Wrath Elixir
	SpellName(60345), -- Armor Piercing
	SpellName(60340), -- Accuracy
	SpellName(60344), -- Expertise
	SpellName(60341), -- Deadly Strikes
	SpellName(60346), -- Lightning Speed
}
local WLK_GElixirs = {
	SpellName(60347), -- Mighty Thoughts
	SpellName(53751), -- Mighty Fortitude
	SpellName(53747), -- Elixir of Spirit
	SpellName(60343), -- Mighty Defense
	SpellName(53763), -- Elixir of Protection
	SpellName(53764), -- Mighty Mageblood
}

local WLK_Good_TBC_Flasks = {
	SpellName(17627), -- [Flask of] Distilled Wisdom
}

local WLK_Good_TBC_BElixirs = {
	SpellName(33721), -- Spellpower Elixir
	SpellName(28491), -- Healing Power
	SpellName(54494), -- Major Agility
	SpellName(28503), -- Major Shadow Power
}

local WLK_Good_TBC_GElixirs = {
	SpellName(39627), -- Elixir of Draenic Wisdom
}

local foods = {
	SpellName(35272), -- Well Fed
	SpellName(44106), -- "Well Fed" from Brewfest
}

local function imerge(dict, ...)
	for arg = 1, select("#", ...) do
		if type(arg) == "table" then
			for _, v in ipairs(arg) do
				dict[#dict+1] = v
			end
		else
			dict[#dict+1] = arg
		end
	end
end

local allFlasks = {}
imerge(allFlasks, TBC_Flasks, WLK_Flasks)

local allBElixirs = {}
imerge(allBElixirs, TBC_BElixirs, WLK_BElixirs)

local allGElixirs = {}
imerge(allGElixirs, TBC_GElixirs, WLK_GElixirs)

local allConsumables = {}
imerge(allConsumables, allFlasks, allBElixirs, allGElixirs, foods)

local inCombat = false

-- Standard Module Methods --------------------------------------------------

-- Is called when the module is loaded
-- @usage internal use only
function Guilder_Monitor:OnModuleLoaded(module)
	if not self.db.profile.enabled then return end
	local id = module.id
end

-- Is called when the module is enabled
-- @usage internal use only
function Guilder_Monitor:OnEnable()
	self:RegisterEvent("READY_CHECK")
	-- enter/leave combat
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	self:RegisterEvent("PLAYER_REGEN_DISABLED")
end

-- Is called when the module is disabled
-- @usage internal use only
function Guilder_Monitor:OnDisable()
	self:UnregisterAllEvents()
end

-- converts the number 'x' to a response string
local function toType(x)
	if x == 0 then return L["No"] end
	if x == 1 then return L["The Burning Crusade"] end
	if x == 2 then return L["Good The Burning Crusade"] end
	if x == 3 then return L["Wraith of the Lich King"] end
end

-- converts the number 'x' to a response string
local function toShortType(x)
	if x == 0 then return L["None"] end
	if x == 1 then return L["TBC"] end
	if x == 2 then return L["Good TBC"] end
	if x == 3 then return L["WotLK"] end
end

-- Is called when a ready check is received
-- @usage internal use only
function Guilder_Monitor:READY_CHECK()
	if self.db.profile.reportOnReadyCheck then
		if self:CheckConsumables() then
			Guilder:Pour("You are ready!", 0, 1, 0)
		else
			Guilder:Pour("You are NOT ready!", 1, 0, 0)
		end
		Guilder:Pour("Ready Check Detected!", 1, 1, 1)
	end
end

-- Is called when the player leaves combat
-- @usage internal use only
function Guilder_Monitor:PLAYER_REGEN_ENABLED()
	local isDead = LibParty:IsDead(myName) or LibRaid:IsDead(myName)
	if self.db.profile.reportOnCombatLeave and (Guilder:IsInParty() or Guilder:IsInRaid()) then
		if isDead then
			Guilder:ShowMessage("|cff00c0ff"..L["Monitor"].."|r: "..L["Checking %s's consumables"]:format(myName))
			-- don't check consumables
			Guilder:ShowMessage("|cffff0000"..L["Not checking for consumables, because you are dead!"].."|r")
			-- do check for dead or offline players
			local dead, offline = checkDeadOrOffline()
			
			if self.db.profile.checkDead and dead ~= "" then
				Guilder:ShowMessage(L["Dead Players"]..":"..dead)
			end
			
			if self.db.profile.checkOffline and offline ~= "" then
				Guilder:ShowMessage(L["Offline Players"]..":"..offline)
			end
			
			Guilder:ShowMessage(L["Checking complete"])
			else
			self:CheckConsumables()
		end
	end
end

-- Is called when the player enters combat
-- @usage internal use only
function Guilder_Monitor:PLAYER_REGEN_DISABLED()
	if Guilder:IsInCombat() then return end
	if self.db.profile.reportOnCombatEnter and (Guilder:IsInParty() or Guilder:IsInRaid()) then
		self:CheckConsumables()
	end
end

-- Is called when the user opens the configuration screen
-- @usage internal use only
function Guilder_Monitor:UpdateConfigOptions()
	local party = (Guilder:IsInParty() and LibParty:IsPartyLeader(myName))
	local raid = (Guilder:IsInRaid() and (LibRaid:IsRaidLeader(myName) or LibRaid:IsRaidAssist(myName)))
	Guilder.options.args.modules.args.Monitor.args.control.args.CheckRaid.disabled = not (party or raid)
end

-- Is called when the module receives a message
-- @usage internal use only
function Guilder_Monitor:CommHandler(distribution, sender, command, ...)
	-- if module is disabled, do nothing
	if not self.db.profile.enabled then
		return
	end
	-- only handle comms if they are not from yourself
	if sender == myName then return end
	local t = {}
	local n = select('#', ...)
	-- Query: Get Ready Status
	if command == "QGRS" then
		Guilder:SendComm(self.id, "WHISPER", sender, "RGRS", whichConsumables())
	end
	-- Response: Get Ready Status
	if command == "RGRS" then
		if n ~= 4 then
			if Guilder.db.profile.showDebug then
				Guilder:ShowMessage("RGRS was expecting 4 parameters!")
			end
			return
		end
		local Flask, BElixir, GElixir, Food = select(1, ...), select(2, ...), select(3, ...), select(4, ...)
		local ready = (Flask > 1 or (BElixir > 1 and GElixir > 1)) and Food > 0
		local msg = ""
		local result
		local addComma = false
		if Flask > 0 then
			if self.db.profile.checkFlasks and Flask < self.db.profile.minFlask then
				msg = L["F"]..": "..toShortType(Flask)
				addComma = true
			end
		else
			if self.db.profile.checkElixirs and BElixir < self.db.profile.minBElixir then
				msg = L["BE"]..": "..toShortType(BElixir)
				addComma = true
			end
			if self.db.profile.checkElixirs and GElixir < self.db.profile.minGElixir then
				if addComma then msg = ", "..msg end
				msg = msg..L["GE"]..": "..toShortType(GElixir)
				addComma = true
			end
		end
		if Food > 0 then
			if self.db.profile.checkFood and Food < self.db.profile.minFood then
				if addComma then msg = ", "..msg end
				msg = msg..L["Not Well Fed"]
			end
		end
		if ready then
			result = "|cff00ff00"..L["ready"].."|r"
		else
			result = "|cffff0000"..L["not ready"].."|r"
		end
		msg = L["%s is %s"]:format(Guilder:GetClassColorString(sender), result).." "..msg
		Guilder:ShowMessage(msg)
	end
end

-- General functions --------------------------------------------------------

-- Checks which consumables a player has
-- The types for Flasks and Elixirs are:
--     0 - None found
--     1 - The Burning Crusade
--     2 - The Burning Crusade, but considered good for Wraith of the Lich King
--     3 - Wraith of the Lich King
-- The types for Food are:
--     0 - player is not Well Fed
--     1 - player is Well Fed
-- @param nil
-- @usage Flask, BElixir, GElixir, Food = whichConsumables()
-- @returns Flask type of flask the player is using
-- BElixir type of battle elixir the player is using
-- GElixir type of guardian elixir the player is using
-- Food type of food the player is using
local function whichConsumables()
	local whichFlask = 0
	local whichBElixir = 0
	local whichGElixir = 0
	local whichFood = 0
	
	-- scan through buffs and compare
	for i = 1, 64 do
		local buffName = UnitBuff(myName, i)
		local name
		if buffName then
			-- check the Flasks
			for k, v in ipairs(TBC_Flasks) do
				if TBC_Flasks[k] == buffName then whichFlask = 1 end
			end
			for k, v in ipairs(WLK_Good_TBC_Flasks) do
				if WLK_Good_TBC_Flasks[k] == buffName then whichFlask = 2 end
			end
			for k, v in ipairs(WLK_Flasks) do
				if WLK_Flasks[k] == buffName then whichFlask = 3 end				
			end
			-- check the Battle Elixirs
			for k, v in ipairs(TBC_BElixirs) do
				if TBC_BElixirs[k] == buffName then whichBElixir = 1 end
			end
			for k, v in ipairs(WLK_Good_TBC_BElixirs) do
				if WLK_Good_TBC_BElixirs[k] == buffName then whichBElixir = 2 end
			end
			for k, v in ipairs(WLK_BElixirs) do
				if WLK_BElixirs[s] == buffName then whichBElixir = 3 end				
			end
			-- check the Guardian Elixirs
			for k, v in ipairs(TBC_GElixirs) do
				if TBC_GElixirs[k] == buffName then whichGElixir = 1 end
			end
			for k, v in ipairs(WLK_Good_TBC_GElixirs) do
				if WLK_Good_TBC_GElixirs[k] == buffName then whichGElixir = 2 end
			end
			for k, v in ipairs(WLK_GElixirs) do
				if WLK_GElixirs[k] == buffName then whichGElixir = 3 end				
			end
			-- check the Food
			for k, v in ipairs(foods) do
				if foods[k] == buffName then whichFood = 1 end				
			end
		end
	end

	return whichFlask, whichBElixir, whichGElixir, whichFood
end

-- Checks if players are dead or offline
-- @usage dead, offline = checkDeadOrOffline()
-- @returns dead a string with all the players that are currently dead
-- offline a string with all the players that are currently offline
local function checkDeadOrOffline()
	if not (Guilder:IsInParty() or Guilder:IsInRaid()) then return end
	local f
	local dead, offline = "", ""
	-- since both LibParty and LibRaid use the same GetIteration and IsDead and IsOnline,
	-- we can use f to represent either, and test once :)
	if Guilder:IsInParty() then f = LibParty end
	if Guilder:IsInRaid() then f = LibRaid end
	for player in f:GetIterator(true) do
		if f:IsDead(player) then dead = dead.." "..player end
		if not f:IsOnline(player) then offline = offline.." "..player end
	end
	return dead, offline
end

-- Checks and reports the status of the player's consumables
-- @usage internal use only
function Guilder_Monitor:CheckConsumables()
	if not self.db.profile.enabled then
		Guilder:ShowMessage(L["%s is disabled"]:format("|cff00c0ff"..L["Monitor"].."|r"))
		return
	end
	-- can only check dead or offline players if we're in a group
	if Guilder:IsInParty() or Guilder:IsInRaid() then
		local dead, offline = checkDeadOrOffline()
		-- report dead players
		if self.db.profile.checkDead and dead ~= "" then
			Guilder:ShowMessage(L["Dead Players"]..":"..dead)
		end
		-- report offline players
		if self.db.profile.checkOffline and offline ~= "" then
			Guilder:ShowMessage(L["Offline Players"]..":"..offline)
		end
	end
	Guilder:ShowMessage("|cff00c0ff"..L["Monitor"].."|r: "..L["Checking %s's consumables"]:format(myName))

	local Flask, BElixir, GElixir, Food = whichConsumables()
	local haveFlask = (Flask > 0)
	local haveBElixir = (BElixir > 0)
	local haveGElixir = (GElixir > 0)
	local haveFood = (Food > 0)
	local buffExpiring = false

	if self.db.profile.checkFlasks and self.db.profile.minFlask > 0 and not (haveBElixir and haveGElixir) then
		Guilder:ShowMessage(L["%s flask found"]:format(toType(Flask)))
	end
	if self.db.profile.checkElixirs and self.db.profile.minElixir > 0 and not haveFlask then 
		Guilder:ShowMessage(L["%s battle elixir found"]:format(toType(BElixir)))
		Guilder:ShowMessage(L["%s guardian elixir found"]:format(toType(GElixir)))
	end
	if self.db.profile.checkFood and self.db.profile.minFood > 0 then
		if haveFood then
			Guilder:ShowMessage(L["Well fed"])
		else
			Guilder:ShowMessage(L["Not well fed"])
		end
	end

	if self.db.profile.checkExpiring then
		local timeLimit = self.db.profile.expiryLimit
		for i = 1, 64 do
			-- buffName, buffRank, buffIcon, buffCount, buffDispelType, buffDuration, buffExpires, buffCaster, buffIsStealable
			local buffName, _, _, _, _, buffDuration, buffExpires, _, _ = UnitBuff(myName, i)
			if buffName and buffDuration then
				for i,v in ipairs(allConsumables) do
					if allConsumables[i] == buffName then
						remainingTime = tonumber(format("%.2f", -1 * (GetTime() - buffExpires) / 60)) or 0
						if remainingTime < timeLimit then
							Guilder:ShowMessage(L["%s will expire in less than %.0f minutes"]:format(buffName, timeLimit))
							buffExpiring = true
						end
					end
				end
			end
		end
	end
	-- determine if we are ready
	local ready = (Flask >= self.db.profile.minFlask or 
		(BElixir >= self.db.profile.minElixir and GElixir >= self.db.profile.minElixir) and
		(Food >= self.db.profile.minFood) and not buffExpiring)
	if ready then
		Guilder:ShowMessage("|cff00ff00"..L["No issues found!"].."|r")
	else
		Guilder:ShowMessage("|cffff0000"..L["Some issues found!"].."|r")
	end
	-- done checking, so tell the user
	Guilder:ShowMessage(L["Checking complete"])
	return ready
end

-- Sends a request to all players with the addon querying their consumables
-- Reports its findings to the player
-- @usage internal use only
function Guilder_Monitor:CheckGroupConsumables()
	if not self.db.profile.enabled then
		Guilder:ShowMessage(L["%s is disabled"]:format("|cff00c0ff"..L["Monitor"].."|r"))
		return
	end
	Guilder:ShowMessage("|cff00c0ff"..L["Monitor"].."|r: "..L["Checking group's consumables"])
	-- check for dead or offline players
	local dead, offline = checkDeadOrOffline()
	-- dead players
	if self.db.profile.checkDead and dead ~= "" then
		Guilder:ShowMessage(L["Dead Players"]..":"..dead)
	end
	-- offline players
	if self.db.profile.checkOffline and offline ~= "" then
		Guilder:ShowMessage(L["Offline Players"]..":"..offline)
	end
	-- query players for their status
	if Guilder:IsInParty() and LibParty:IsPartyLeader(myName) then
		Guilder:SendComm(self.id, "PARTY", myName, "QGRS")
	end
	if Guilder:IsInRaid() and (Guilder:IsOfficer(myName) or Guilder:IsOfficerAlt(myName) or LibRaid:IsRaidLeader(myName) or LibRaid:IsRaidAssist(myName)) then
		Guilder:SendComm(self.id, "RAID", myName, "QGRS")
	end
end
